@@ -4,7 +4,8 @@ import os
 from abc import ABC, abstractmethod
 from collections import defaultdict
 from collections.abc import Iterator, MutableMapping
-from typing import TYPE_CHECKING, Any, Callable, Optional, cast
+from itertools import repeat
+from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast
 
 import attrs
 from fsspec import Callback
@@ -17,6 +18,8 @@ from dvc_data.hashfile.meta import Meta
 from dvc_data.hashfile.tree import Tree
 
 if TYPE_CHECKING:
+    from collections.abc import Iterable
+
     from dvc_objects.fs.base import FileSystem
 
     from dvc_data.hashfile.db import HashFileDB
@@ -240,13 +243,15 @@ class ObjectStorage(Storage):
         finally:
             self.index.commit()
 
-    def bulk_exists(
+    def bulk_exists(  # noqa: C901, PLR0912
         self,
         entries: list["DataIndexEntry"],
         refresh: bool = False,
         jobs: Optional[int] = None,
         callback: "Callback" = DEFAULT_CALLBACK,
     ) -> dict["DataIndexEntry", bool]:
+        from .build import build_entry
+
         entries_with_hash = [e for e in entries if e.hash_info]
         entries_without_hash = [e for e in entries if not e.hash_info]
         results = dict.fromkeys(entries_without_hash, False)
@@ -267,41 +272,47 @@ class ObjectStorage(Storage):
                 results[entry] = exists
             return results
 
-        entry_map: dict[str, DataIndexEntry] = {
-            self.get(entry)[1]: entry for entry in entries_with_hash
-        }
+        path_to_entries: dict[str, list[DataIndexEntry]] = defaultdict(list)
+        for entry in entries_with_hash:
+            _, path = self.get(entry)
+            path_to_entries[path].append(entry)
 
+        info_results: Union[
+            Iterable[Union[Exception, Optional[dict[str, Any]]]], None
+        ] = None
         try:
             self.fs.ls(self.odb.path)  # check for fs access
-        except FileNotFoundError:
+        except FileNotFoundError as exc:
+            info_results = repeat(exc, len(path_to_entries))
+            callback.relative_update(len(entries_with_hash))
+        except NotImplementedError:
+            # some filesystems don't implement ls
             pass
+        if info_results is None:
+            info_results = self.fs.info(
+                list(path_to_entries),
+                batch_size=jobs,
+                return_exceptions=True,
+                callback=callback,
+            )
 
-        info_results = self.fs.info(
-            list(entry_map.keys()),
-            batch_size=jobs,
-            return_exceptions=True,
-            callback=callback,
-        )
-
-        for (path, entry), info in zip(entry_map.items(), info_results):
+        for (path, _entries), info in zip(path_to_entries.items(), info_results):
+            if isinstance(info, Exception) and not isinstance(info, FileNotFoundError):
+                raise info
+            assert _entries
+            entry = _entries[0]
             assert entry.hash_info  # built from entries_with_hash
             value = cast("str", entry.hash_info.value)
             key = self.odb._oid_parts(value)
-
-            if isinstance(info, FileNotFoundError) or info is None:
+            exists = info is not None and not isinstance(info, FileNotFoundError)
+            if exists:
+                self.index[key] = build_entry(path, self.fs, info=info)
+            else:
                 self.index.pop(key, None)
-                results[entry] = False
-                continue
-            if isinstance(info, Exception):
-                raise info
-
-            from .build import build_entry
-
-            built_entry = build_entry(path, self.fs, info=info)
-            self.index[key] = built_entry
-            results[entry] = True
+            results.update(dict.fromkeys(_entries, exists))
 
         if self.index is not None:
+            logger.debug("Committing index results")
             self.index.commit()
 
         return results
